// This file is part of PUMA.
// Copyright (C) 1999-2003  The PUMA developer team.
//
// This program is free software;  you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
// MA  02111-1307  USA

#ifndef __Config_h__
#define __Config_h__

/** \file
 *  Configuration information from file or command line. */

/** \page configuration Configuration Options
 *
 * Several aspects of the functionality of PUMA can be conﬁgured using command line 
 * options or a configuration file. The following configuration options are understood 
 * by the library.
 * 
 * ### File Handling Options
 * 
 * | Option | Description
 * |----------------------|-----------------------
 * | -p, --path PATH | Add a project source directory
 * | -d, --dest PATH | Add a destination folder for modified sources
 * | -w, --write-protected PATH | Add a write protected project directory
 * | -e, --extension STRING | Set the extension for source files
 * | -s, --suffix STRING | Set the suffix for output files
 * | --config FILE | Load configuration options from a file
 * | --system-config FILE | Load low precedence configuration options from a file
 * | --save-new-suffix | Append new suffix on existing files when saving
 * | --save-overwrite | Overwrite existing files when saving
 * | --save-rename-old | Rename existing files when saving
 * 
 * ### Preprocessor Options
 * 
 * | Option | Description
 * |----------------------|-----------------------
 * | -A PREDICATE(ANSWER) | Assert a preprocessor predicate, e.g. -Asystem(linux)
 * | -D NAME[=BODY] | Define a preprocessor macro, e.g. -DSYSTEM=linux or -D DEBUG
 * | -U NAME | Undefine a preprocessor macro, e.g. -U SYSTEM
 * | -I PATH | Add a non-system include directory
 * | --isystem PATH | Add a system include directory
 * | --include FILE | Include the given file at the beginning of each source file
 * | --lock-macro NAME[=BODY] | Define an immutable preprocessor macro
 * | --inhibit-macro NAME | Add the name of a macro whose definition will be ignored
 * 
 * ### General %Parser Options
 * 
 * | Option | Description
 * |----------------------|-----------------------
 * | --lang-c | Set input language to C
 * | --lang-ec++ | Set input language to Embedded C++
 * | --lang-c++ | Set input language to C++
 * | --lang-ac++ | Set input language AspectC++
 * | --size-type TYPE | Set the internal type for size_t
 * | --ptrdiff-type TYPE | Set the internal type for ptrdiff_t
 * | --target TRIPLE | Set target triple which determines sizes and alignments for built-in types
 * | --builtin-type-traits | Enable built-in type traits
 * | --verbose-errors | Enable verbose error messages
 * | --match-expr | Enable match expression language extensions
 * 
 * ### Extended %Parser Options
 * 
 * | Option | Description
 * |----------------------|-----------------------
 * | --skip-bodies-all | Don't parse function bodies
 * | --skip-bodies-tpl | Don't parse function bodies of templates
 * | --skip-bodies-non-prj | Don't parse function bodies in non-project files
 * | --skip-bodies-non-prim | Don't parse function bodies in non-primary files
 * | --pseudo-instances | Simplified template instantiation scheme (deprecated)
 * | --inst-fct-bodies | Enable function template body instantiation
 * | --template-depth DEPTH | Set the maximum instantiation depth for templates
 * 
 * ### VisualC++ Extension Options
 * 
 * | Option | Description
 * |----------------------|-----------------------
 * | --vc | Enable VisualC++ language extensions
 * | --import-handler FILE | Set a handler for resolving \#import directives
 * 
 * ### GNU C/C++ Extension Options
 * 
 * | Option | Description
 * |----------------------|-----------------------
 * | --gnu [VERSION] | Enable all GNU C/C++ language extensions (optionally version dependent)
 * | --gnu-nested-fct | Enable GNU C/C++ nested functions
 * | --gnu-condition-scope | Enable GNU C/C++ condition scope
 * | --gnu-param-decl | Enable GNU C/C++ parameter declarator
 * | --gnu-fct-decl | Enable GNU C/C++ function declarator
 * | --gnu-param-scope | Enable GNU C/C++ function parameter scope
 * | --gnu-default-args | Enable GNU C/C++ function default arguments
 * | --gnu-extended-asm | Enable GNU C/C++ extended asm syntax
 * | --gnu-extended-expr | Enable GNU C/C++ extended expressions
 * | --gnu-long-long | Enable GNU C/C++ long long type
 * | --gnu-int128 | Enable GNU type __int128 extension
 * | --gnu-name-scope | Enable GNU C/C++ name scope
 * | --gnu-implicit-int | Enable GNU implicit int extension
 * | --gnu-fct-attribute | Enable GNU C/C++ function attributes
 * | --gnu-if-then-expr | Enable GNU C/C++ if-then expression syntax
 * | --gnu-std-hack | Enable GNU C/C++ implicit namespace std hack
 * | --gnu-friend-injection | Enable GNU friend injection extension
 * | --gnu-pic24 | Enable GNU PIC24 and dsPIC extension
 * 
 * ### C++1x Extension Options
 * 
 * | Option | Description
 * |----------------------|-----------------------
 * | --c++1x | Enable all C++1x language extensions
 * | --c++1x-static-assert | Enable C++1x static assertions
 * 
 * ### Configuration Files
 * 
 * All command line options can also be specified in a configuration file. 
 * Each option in the configuration file has to start on a new line.
 * 
 * \code{.py}
 * -D i386
 * -D linux
 * -I /usr/include
 * \endcode
 * 
 * Lines beginning with # are interpreted as comments and will be ignored.
 * 
 * \code{.py}
 * ### defines
 * -D i386
 * -D linux
 * ### includes
 * -I /usr/include
 * -I /usr/local/include
 * \endcode
 *
 * Arguments of options containing spaces have to be enclosed in double-quotes. 
 * Double-quotes in the argument have to be escaped.
 * 
 * \code{.py}
 * -D __PTRDIFF_TYPE__=int
 * -D "__SIZE_TYPE__=unsigned int"
 * -D "__VERSION__=\"4.1.0 (Linux)\""
 * \endcode
 * 
 * All occurrences of ${Name} in the configuration file are interpreted as environment 
 * variables and replaced by their values, or by nothing if the variable is not defined. 
 * To avoid variable replacement, $ has to be escaped.
 * 
 * \code{.py}
 * -I ${LIBDIR}/include
 * -D OS_STR=\"${OSTYPE}\"
 * ### same as: #define OS_STR "linux"
 * -D OS_VAR=\"\${OSTYPE}\"
 * ### same as: #define OS_VAR "${OSTYPE}"
 * \endcode
 */

#include "Puma/Flags.h"
#include "Puma/ConfOption.h"
#include "Puma/ErrorStream.h"
#include "Puma/OptsParser.h"
#include "Puma/StrCol.h"
#include <iostream>
#include <string>
#include <vector>
#include <map>

namespace Puma {


/** \class Config Config.h Puma/Config.h
 *  PUMA configuration information. Can read the
 *  configuration from file or command line. */
class Config {
  typedef std::vector<const ConfOption *> OptionList;
  typedef std::map<unsigned int, const ConfOption*> OptionMap;

  Flags _flags;
  ErrorStream &_err;
  OptionList _sysopts;
  OptionList _useropts;
  OptionList _sysincls;  
  OptionMap _optnamemap;
  bool _system;

  static unsigned int OPT_I_HASH;
  static unsigned int OPT_ISYSTEM_HASH;

public:
  enum OPTIONS {
    PRE_ASSERT = 1,
    PRE_DEFINE,
    PRE_UNDEFINE,
    PRE_LOCK_MACRO,
    PRE_INCLUDE,
    CFG_FILE,
    SYS_CFG_FILE,
    PROJ_PATH,
    PROJ_DESTINATION,
    SET_SUFFIX,
    SET_EXTENSION,
    SET_WRITE_PROTECTED,
    GNU,
    SET_OPTION,
    SET_OPTION_ARG
  };

public:
  Config (ErrorStream &);
  Config (const Config &);
  ~Config ();

  // read config file
  Config (ErrorStream &, const char *);
  // read command line
  Config (ErrorStream &, int &, char **&);
  // read config file and command line
  Config (ErrorStream &, const char *, int &, char **&);

  // join with another configuration
  void Join (const Config &);

  // evaluate file, if file not given read from $PUMA_CONFIG
  // or /etc/puma.config
  void Read (const char * = (const char *)0);
  // evaluate command line
  void Read (int &, char **&);

  // add options to configuration
  void Add (const ConfOption *);
  void Add (const char *, bool system = false);
  void Add (const char *, const char *, bool system = false);
  void Add (const char *, const char *, const char *, bool system = false);

  // remove option from the configuration
  void Remove (const ConfOption *);

  // get options
  unsigned Options () const;
  const ConfOption *Option (unsigned) const;
  const ConfOption *Option (const char *) const;

  // check if a config file is given on command line
  bool CustomConfigFile (int argc, char **argv);

  // check if a system config file is given on command line
  // for lower priority options
  bool CustomSystemConfigFile (int argc, char **argv);

  Flags FlagTable () const;

  // print the options
  void PrintOptions(std::ostream& os) const;
  // print a single option
  void PrintOption(OptsParser::Option& o, std::ostream& os) const;

  // replace environment variables of the form ${ENVVAR}
  std::string& ReplaceEnvVars(std::string& s) const;

private:
  bool Valid (const char *, const char * = (const char *)0) const;

  bool Process (OptsParser&, bool unescape = false);
  bool ProcessAssert (const std::string&);
  bool ProcessDefine (const std::string&);
  bool ProcessLockMacro (const std::string&);
  bool ProcessUndefine (const std::string&);
  bool ProcessConfigFile (const std::string&,const std::string&);
  bool ProcessSystemConfigFile (const std::string&,const std::string&);
  bool ProcessArgument (const std::string&,const std::string&);
  bool ProcessOptionalArgument (const std::string&,const std::string&);
  bool ProcessPathArgument (const std::string&,const std::string&);

  std::string Unescape(const std::string&);

  void Remove (const ConfOption *, OptionList& options);
};

inline Config::Config (ErrorStream &e) : _err (e), _system (false)
 { }
inline Config::Config (const Config &c) : _err (c._err), _system (false)
 { Join (c); }

inline Config::Config (ErrorStream &e, const char *file) : _err (e), _system (false)
 { Read (file); }
inline Config::Config (ErrorStream &e, int &argc, char **&argv) : _err (e), _system (false)
 { Read (argc, argv); }
inline Config::Config (ErrorStream &e, const char *file, int &argc, char **&argv) : _err (e), _system (false)
 { Read (file); Read (argc, argv); }

inline Flags Config::FlagTable () const
 { return _flags; }

inline void Config::Add (const char *n, bool system)
 { Add (new ConfOption (n, system)); }
inline void Config::Add (const char *n, const char *a1, bool system)
 { Add (new ConfOption (n, a1, system)); }
inline void Config::Add (const char *n, const char *a1, const char *a2, bool system)
 { Add (new ConfOption (n, a1, a2, system)); }


} // namespace Puma

#endif /* __Config_h__ */
